/* SAMHAIN file system integrity testing                                   */
/* Copyright (C) 1999 Rainer Wichmann                                      */
/*                                                                         */
/*  This program is free software; you can redistribute it                 */
/*  and/or modify                                                          */
/*  it under the terms of the GNU General Public License as                */
/*  published by                                                           */
/*  the Free Software Foundation; either version 2 of the License, or      */
/*  (at your option) any later version.                                    */
/*                                                                         */
/*  This program is distributed in the hope that it will be useful,        */
/*  but WITHOUT ANY WARRANTY; without even the implied warranty of         */
/*  MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the          */
/*  GNU General Public License for more details.                           */
/*                                                                         */
/*  You should have received a copy of the GNU General Public License      */
/*  along with this program; if not, write to the Free Software            */
/*  Foundation, Inc., 675 Mass Ave, Cambridge, MA 02139, USA.              */

#include "config_xor.h"

/* #define SH_PROFILE 1 */

#include <stdio.h>
#include <stdlib.h>
#include <string.h>
#include <ctype.h>


#ifdef HAVE_MEMORY_H
#include <memory.h>
#endif

#ifdef  HAVE_UNISTD_H
#include <errno.h>
#include <signal.h>
#include <pwd.h>
#include <grp.h>
#include <sys/types.h>
#include <sys/stat.h>
#include <sys/resource.h>
#include <fcntl.h>
#include <unistd.h>

/*********************
#ifdef HAVE_SYS_VFS_H
#include <sys/vfs.h>
#endif
**********************/
#endif

#if TIME_WITH_SYS_TIME
#include <sys/time.h>
#include <time.h>
#else
#if HAVE_SYS_TIME_H
#include <sys/time.h>
#else
#include <time.h>
#endif
#endif

#ifdef HAVE_SYS_SELECT_H
#include <sys/select.h>
#endif

#ifndef FD_SET
#define NFDBITS         32
#define FD_SET(n, p)    ((p)->fds_bits[(n)/NFDBITS] |= (1 << ((n) % NFDBITS)))
#define FD_CLR(n, p)    ((p)->fds_bits[(n)/NFDBITS] &= ~(1 << ((n) % NFDBITS)))
#define FD_ISSET(n, p)  ((p)->fds_bits[(n)/NFDBITS] & (1 << ((n) % NFDBITS)))
#endif /* !FD_SET */
#ifndef FD_SETSIZE
#define FD_SETSIZE      32
#endif
#ifndef FD_ZERO
#define FD_ZERO(p)      memset((char *)(p), '\0', sizeof(*(p)))
#endif


#if defined(HAVE_MLOCK) && !defined(HAVE_BROKEN_MLOCK)
#include <sys/mman.h>
#endif

#include "samhain.h"
#include "sh_error.h"
#include "sh_unix.h"
#include "sh_utils.h"
#include "sh_mem.h"
#include "sh_hash.h"
#include "sh_tools.h"
#include "sh_tiger.h"


#ifndef HAVE_LSTAT
#define lstat   stat
#endif 

#undef  FIL__
#define FIL__  _("sh_unix.c")

unsigned long mask_USER0        = MASK_USER_;
unsigned long mask_USER1        = MASK_USER_;
unsigned long mask_ALLIGNORE    = MASK_ALLIGNORE_;
unsigned long mask_ATTRIBUTES   = MASK_ATTRIBUTES_;
unsigned long mask_LOGFILES     = MASK_LOGFILES_;
unsigned long mask_LOGGROW      = MASK_LOGFILES_;
unsigned long mask_READONLY     = MASK_READONLY_;
unsigned long mask_NOIGNORE     = MASK_NOIGNORE_;


extern char **environ;

int sh_unix_maskreset()
{
  mask_USER0        = MASK_USER_;
  mask_USER1        = MASK_USER_;
  mask_ALLIGNORE    = MASK_ALLIGNORE_;
  mask_ATTRIBUTES   = MASK_ATTRIBUTES_;
  mask_LOGFILES     = MASK_LOGFILES_;
  mask_LOGGROW      = MASK_LOGFILES_;
  mask_READONLY     = MASK_READONLY_;
  mask_NOIGNORE     = MASK_NOIGNORE_;
  return 0;
}

#ifdef SYS_SIGLIST_DECLARED

extern const char * const sys_siglist[];

#else

char * sh_unix_siglist (int signum)
{

  switch (signum)
    {
#ifdef SIGHUP
    case SIGHUP: 
      return _("Hangup");
#endif
#ifdef SIGINT
    case SIGINT: 
      return _("Interrupt");
#endif
#ifdef SIGQUIT
    case SIGQUIT: 
      return _("Quit");
#endif
#ifdef SIGILL
    case SIGILL: 
      return _("Illegal instruction");
#endif
#ifdef SIGTRAP
    case SIGTRAP: 
      return _("Trace/breakpoint trap");
#endif
#ifdef SIGABRT
    case SIGABRT: 
      return _("IOT trap/Abort");
#endif
#ifdef SIGBUS
    case SIGBUS: 
      return _("Bus error");
#endif
#ifdef SIGFPE
    case SIGFPE: 
      return _("Floating point exception");
#endif
#ifdef SIGUSR1
    case SIGUSR1: 
      return _("User defined signal 1");
#endif
#ifdef SIGSEGV
    case SIGSEGV: 
      return _("Segmentation fault");
#endif
#ifdef SIGUSR2
    case SIGUSR2: 
      return _("User defined signal 2");
#endif
#ifdef SIGPIPE
    case SIGPIPE: 
      return _("Broken pipe");
#endif
#ifdef SIGALRM
    case SIGALRM: 
      return _("Alarm clock");
#endif
#ifdef SIGTERM
    case SIGTERM: 
      return _("Terminated");
#endif
#ifdef SIGSTKFLT
    case SIGSTKFLT: 
      return _("Stack fault");
#endif
#ifdef SIGCHLD
    case SIGCHLD: 
      return _("Child exited");
#endif
#ifdef SIGCONT
    case SIGCONT: 
      return _("Continued");
#endif
#ifdef SIGSTOP
    case SIGSTOP: 
      return _("Stopped");
#endif
#ifdef SIGTSTP
    case SIGTSTP: 
      return _("Stopped");
#endif
#ifdef SIGTTIN
    case SIGTTIN: 
      return _("Stopped (tty input)");
#endif
#ifdef SIGTTOU
    case SIGTTOU: 
      return _("Stopped (tty output)");
#endif
#ifdef SIGURG
    case SIGURG: 
      return _("Urgent condition");
#endif
#ifdef SIGXCPU
    case SIGXCPU: 
      return _("CPU time limit exceeded");
#endif
#ifdef SIGXFSZ
    case SIGXFSZ: 
      return _("File size limit exceeded");
#endif
#ifdef SIGVTALRM
    case SIGVTALRM: 
      return _("Virtual time alarm");
#endif
#ifdef SIGPROF
    case SIGPROF: 
      return _("Profile signal");
#endif
#ifdef SIGWINCH
    case SIGWINCH: 
      return _("Window size changed");
#endif
#ifdef SIGIO
    case SIGIO: 
      return _("Possible I/O");
#endif
#ifdef SIGPWR
    case SIGPWR: 
      return _("Power failure");
#endif
#ifdef SIGUNUSED
    case SIGUNUSED: 
      return _("Unused signal");
#endif
    }
  return _("Unknown");
}

#endif

extern char sh_sig_msg[64];

volatile int immediate_exit_normal = 0;

static
void sh_unix_sigexit (int mysignal
#if defined(SA_SIGINFO) && defined(SI_USER)
		      , siginfo_t * signal_info, void * signal_add
#endif
		      )
{

#if defined(SA_SIGINFO) && defined(SI_USER) && !defined(__linux__) 
  /* On linux (as of 2.2.16) no useful information in 
   * siginfo_t * signal_info.
   * On Solaris, siginfo_t* may be a NULL pointer :-(
   * All in all, quite a few implementations seem to be crap.
   */
  if (signal_info != NULL && signal_info->si_code == SI_USER && 
      mysignal != SIGTERM)
    return;
#endif

  /* block re-entry
   */
  if (immediate_exit_normal > 0)
    return;
  else
    immediate_exit_normal = 1;

#if defined(SA_SIGINFO) && defined(SI_USER) && defined(__linux__)
  signal_info = NULL;
  signal_add  = NULL;
#elif defined(SA_SIGINFO) && defined(SI_USER)
  signal_add  = NULL;
#endif

#ifdef SYS_SIGLIST_DECLARED
  strncpy (sh_sig_msg, sys_siglist[mysignal],     40);
#else
  strncpy (sh_sig_msg, sh_unix_siglist(mysignal), 40);
#endif

  sh_sig_msg[63] = '\0';
  exit(mysignal);
}

volatile int immediate_exit_fast = 0;

static
void sh_unix_sigexit_fast (int mysignal
#if defined(SA_SIGINFO) && defined(SI_USER)
			   , siginfo_t * signal_info, void * signal_add
#endif
			   )
{
  /* On Linux there is no useful information in siginfo_t
   */
#if defined(SA_SIGINFO) && defined(SI_USER) && !defined(__linux__)
  /* On linux (as of 2.2.16) no useful information in 
   * siginfo_t * signal_info.
   * On Solaris, signal_info may be a NULL pointer :-(
   */
  if (signal_info != NULL && signal_info->si_code == SI_USER)
    return;
#endif

#if defined(SA_SIGINFO) && defined(SI_USER) && defined(__linux__)
  signal_info = NULL;
  signal_add  = NULL;
#elif defined(SA_SIGINFO) && defined(SI_USER)
  signal_add  = NULL;
#endif

  if (immediate_exit_fast == 1)
    {
      ++immediate_exit_fast;
      if (skey != NULL)
	memset (skey, '\0', sizeof(sh_key_t));
#ifdef WITH_MESSAGE_QUEUE
      close_ipc ();
#endif
#if !defined(SL_DEBUG)
      _exit(mysignal);
#else
      raise(SIGFPE);
#endif
    }
  else if (immediate_exit_fast != 0)
    {
#if !defined(SL_DEBUG)
      _exit(mysignal);
#else
      raise(SIGFPE);
#endif
    }

  ++immediate_exit_fast;
  
  /* The FPE|BUS|SEGV|ILL signals leave the system in an undefined
   * state, thus it is best to exit immediately.
   */
#ifdef SYS_SIGLIST_DECLARED
  strncpy (sh_sig_msg, sys_siglist[mysignal],     40);
#else
  strncpy (sh_sig_msg, sh_unix_siglist(mysignal), 40);
#endif

#if defined(SL_DEBUG)
  sl_stack_print();
#endif

  /* Try to push out an error message. This might be dangerous, although
   * we have a block around free()/malloc() in the wrapper function.
   */
  sh_error_handle ((-1), FIL__, __LINE__, mysignal, MSG_EXIT_NORMAL, 
		   sh.prg_name, sh_sig_msg);

  if (skey != NULL)
    memset (skey, '\0', sizeof(sh_key_t));
#ifdef WITH_MESSAGE_QUEUE
  close_ipc ();
#endif

#if !defined(SL_DEBUG)
  _exit(mysignal);
#else
  raise(SIGFPE);
#endif
}


static
void sh_unix_sigaction (int mysignal)
{
  sig_raised = 1;
#ifdef SIGUSR1
  if (mysignal == SIGUSR1)
    sig_debug_on          = 1;
#endif
#ifdef SIGUSR2
  if (mysignal == SIGUSR2)
    sig_debug_off         = 1;
#endif
#ifdef SIGHUP
  if (mysignal == SIGHUP)
    sig_config_read_again = 1;
#endif
#ifdef SIGABRT
  if (mysignal == SIGABRT)
    sig_fresh_trail       = 1;
#endif
#ifdef SIGQUIT
  if (mysignal == SIGQUIT)
    sig_terminate         = 1;
#endif

  return;
}

static
void sh_unix_siginstall (void)
{
  struct sigaction act, act_fast, act2, oldact, ignact;

  SL_ENTER(_("sh_unix_siginstall"));

  ignact.sa_handler = SIG_IGN;          /* signal action           */
  sigemptyset( &ignact.sa_mask );       /* set an empty mask       */
  ignact.sa_flags = 0;                  /* init sa_flags           */

#if defined(SA_SIGINFO) && defined(SI_USER)
  act.sa_sigaction = &sh_unix_sigexit;    /* signal action           */
#else
  act.sa_handler   = &sh_unix_sigexit;    /* signal action           */
#endif

  sigfillset ( &act.sa_mask );            /* set a  full mask        */

  /* Block all but deadly signals.
   */
#ifdef SIGILL
  sigdelset  ( &act.sa_mask, SIGILL  );
#endif
#ifndef SL_DEBUG
#ifdef SIGFPE
  sigdelset  ( &act.sa_mask, SIGFPE  );
#endif
#endif
#ifdef SIGSEGV
  sigdelset  ( &act.sa_mask, SIGSEGV );
#endif
#ifdef SIGBUS
  sigdelset  ( &act.sa_mask, SIGBUS  );
#endif

#if defined(SA_SIGINFO) && defined(SI_USER)
  act_fast.sa_sigaction = &sh_unix_sigexit_fast;  /* signal action           */
#else
  act_fast.sa_handler   = &sh_unix_sigexit_fast;  /* signal action           */
#endif

  sigfillset ( &act_fast.sa_mask );               /* set a full mask         */

#ifdef SIGILL
  sigdelset  ( &act_fast.sa_mask, SIGILL  );
#endif
#ifndef SL_DEBUG
#ifdef SIGFPE
  sigdelset  ( &act_fast.sa_mask, SIGFPE  );
#endif
#endif
#ifdef SIGSEGV
  sigdelset  ( &act_fast.sa_mask, SIGSEGV );
#endif
#ifdef SIGBUS
  sigdelset  ( &act_fast.sa_mask, SIGBUS  );
#endif


  /* Use siginfo to verify origin of signal, if possible.
   */
#if defined(SA_SIGINFO) && defined(SI_USER)
  act.sa_flags      = SA_SIGINFO;
  act_fast.sa_flags = SA_SIGINFO;
#else
  act.sa_flags      = 0;
  act_fast.sa_flags = 0;
#endif 

  /* Do not block the signal from being received in its handler ...
   * (is this a good or a bad idea ??).
   */
#if   defined(SA_NOMASK)
  act_fast.sa_flags |= SA_NOMASK;
#elif defined(SA_NODEFER)
  act_fast.sa_flags |= SA_NODEFER;
#endif


  act2.sa_handler = &sh_unix_sigaction;  /* signal action           */
  sigemptyset( &act2.sa_mask );          /* set an empty mask       */
  act2.sa_flags = 0;                     /* init sa_flags           */

  /* signals to control the daemon */

#ifdef SIGHUP
  retry_sigaction(FIL__, __LINE__, SIGHUP,  &act2, &oldact);
#endif
#ifdef SIGABRT
  retry_sigaction(FIL__, __LINE__, SIGABRT, &act2, &oldact);
#endif
#ifdef SIGUSR1
  retry_sigaction(FIL__, __LINE__, SIGUSR1, &act2, &oldact);
#endif
#ifdef SIGUSR2
  retry_sigaction(FIL__, __LINE__, SIGUSR2, &act2, &oldact);
#endif
#ifdef SIGQUIT
  retry_sigaction(FIL__, __LINE__, SIGQUIT, &act2, &oldact);
#endif
#ifdef SIGTERM
  retry_sigaction(FIL__, __LINE__, SIGTERM, &act, &oldact);
#endif

  /* fatal signals that may cause termination */

#ifdef SIGILL
  retry_sigaction(FIL__, __LINE__, SIGILL,  &act_fast, &oldact);
#endif
#ifndef SL_DEBUG
#ifdef SIGFPE
  retry_sigaction(FIL__, __LINE__, SIGFPE,  &act_fast, &oldact);
#endif
#endif
#ifdef SIGSEGV
  retry_sigaction(FIL__, __LINE__, SIGSEGV, &act_fast, &oldact);
#endif
#ifdef SIGBUS
  retry_sigaction(FIL__, __LINE__, SIGBUS,  &act_fast, &oldact);
#endif

  /* other signals  */

#ifdef SIGINT
  retry_sigaction(FIL__, __LINE__, SIGINT,  &act, &oldact);
#endif
#ifdef SIGPIPE
  retry_sigaction(FIL__, __LINE__, SIGPIPE, &act, &oldact);
#endif
#ifdef SIGALRM
  retry_sigaction(FIL__, __LINE__, SIGALRM, &ignact, &oldact);
#endif
#ifdef SIGTSTP
  retry_sigaction(FIL__, __LINE__, SIGTSTP, &ignact, &oldact);
#endif
#ifdef SIGTTIN
  retry_sigaction(FIL__, __LINE__, SIGTTIN, &act, &oldact);
#endif
#ifdef SIGTTOU
  retry_sigaction(FIL__, __LINE__, SIGTTOU, &ignact, &oldact);
#endif

#ifdef SIGTRAP
  retry_sigaction(FIL__, __LINE__, SIGTRAP, &act, &oldact);
#endif
#ifdef SIGEMT
  retry_sigaction(FIL__, __LINE__, SIGEMT,  &ignact, &oldact);
#endif
#ifdef SIGSTKFLT
  retry_sigaction(FIL__, __LINE__, SIGSTKFLT,&act, &oldact);
#endif
#ifdef SIGIO
  retry_sigaction(FIL__, __LINE__, SIGIO,   &ignact, &oldact);
#endif
#ifdef SIGPOLL
  retry_sigaction(FIL__, __LINE__, SIGPOLL, &ignact, &oldact);
#endif
#ifdef SIGXCPU
  retry_sigaction(FIL__, __LINE__, SIGXCPU, &act, &oldact);
#endif
#ifdef SIGXFSZ
  retry_sigaction(FIL__, __LINE__, SIGXFSZ, &act, &oldact);
#endif
#if defined(SIGVTALRM) && !defined(SH_PROFILE)
  retry_sigaction(FIL__, __LINE__, SIGVTALRM, &ignact, &oldact);
#endif
#if defined(SIGPROF) && !defined(SH_PROFILE)
  retry_sigaction(FIL__, __LINE__, SIGPROF, &ignact, &oldact);
#endif
#ifdef SIGPWR
  retry_sigaction(FIL__, __LINE__, SIGPWR,  &act, &oldact);
#endif

#ifdef SIGLOST
  retry_sigaction(FIL__, __LINE__, SIGLOST,  &ignact, &oldact);
#endif
#ifdef SIGUNUSED
  retry_sigaction(FIL__, __LINE__, SIGUNUSED,  &ignact, &oldact);
#endif

  SL_RET0(_("sh_unix_siginstall"));
}

/* ---------------------------------------------------------------- */

/* checksum the own binary
 */
int sh_unix_self_hash (char * c)
{
  SL_ENTER(_("sh_unix_self_hash"));

  if (c == NULL)
    SL_RETURN((0), _("sh_unix_self_hash"));

  sl_strlcpy(sh.exec.path, c, SH_PATHBUF);

  sl_strlcpy(sh.exec.hash,
	     sh_tiger_hash (c, TIGER_FILE, 0), 
	     KEY_LEN+1);
  if (0 == sl_strcmp(sh.exec.hash,
		     _("000000000000000000000000000000000000000000000000")
		     ))
    {
      dlog(1, FIL__, __LINE__, 
	   _("Could not checksum my own executable because of the\nfollowing error: %s\nIf this is a permission problem, you need to change file permissions\nto make the file readable for the effective UID: %d\n"), 
	   sl_get_errmsg(), (int) sl_ret_euid());
      sh_error_handle ((-1), FIL__, __LINE__, EACCES, MSG_NOACCESS,
		       (long) sh.real.uid, c);
      aud_exit (FIL__, __LINE__, EXIT_FAILURE);
    }
  SL_RETURN((0), _("sh_unix_self_hash"));
}

int sh_unix_self_check ()
{
  SL_ENTER(_("sh_unix_self_check"));
  if (sh.exec.path == NULL || sh.exec.path[0] == '\0')
    SL_RETURN((0), _("sh_unix_self_check"));

  if (0 == sl_strncmp(sh.exec.hash, 
		      sh_tiger_hash (sh.exec.path, TIGER_FILE, 0), 
		      KEY_LEN))
    SL_RETURN((0), _("sh_unix_self_check"));

 
    dlog(1, FIL__, __LINE__, 
	 _("The checksum of the executable: %s has changed since startup.\n"),
	 sh.exec.path);
    sh_error_handle ((-1), FIL__, __LINE__, EACCES, MSG_E_AUTH,
		     sh.exec.path);
    SL_RETURN((-1), _("sh_unix_self_check"));
}


/* ---------------------------------------------------------------- */
  



/* added    Tue Feb 22 10:36:44 NFT 2000 Rainer Wichmann            */
static int tf_add_trusted_user_int(char * c)
{
  register struct passwd *          w;
  int                           count;
  long                     pwid  = -1;

  SL_ENTER(_("tf_add_trusted_user_int"));

  /* First check for a user name.
   */
  if ((w = getpwnam(c)) != NULL && ((pwid = w->pw_uid) > 0))
    goto succe;
	
  /* Failed, so check for a numerical value.
   */
  pwid = strtol(c, (char **)NULL, 10);
  if (pwid > 0 && pwid < 65535)
    goto succe;
      
  sh_error_handle ((-1), FIL__, __LINE__, EINVAL, MSG_EINVALS, 
		   _("add trusted user"), c);
  SL_RETURN((-1), _("tf_add_trusted_user_int"));

 succe:
  count = sl_trust_add_user(pwid);
  SL_RETURN((count), _("tf_add_trusted_user_int"));
}

int tf_add_trusted_user(char * c)
{
  int    i;
  char * q;
  char * p = sh_util_strdup (c);
  
  SL_ENTER(_("tf_add_trusted_user"));

  q = strtok(p, ", ");
  if (!q)
    {
      SH_FREE(p);
      SL_RETURN((-1), _("tf_add_trusted_user"));
    }
  while (q)
    {
      i = tf_add_trusted_user_int(q);
      if (SL_ISERROR(i))
	{
	  SH_FREE(p);
	  SL_RETURN((i), _("tf_add_trusted_user"));
	}
      q = strtok(NULL, ", ");
    }
  SH_FREE(p);
  SL_RETURN((0), _("tf_add_trusted_user"));
}

extern long   sl_trust_baduid();
extern long   sl_trust_badgid();

int tf_trust_check (char * file, int mode)
{
  char * tmp;
  char * tmp2;
  char * p;
  int    status;
  int    level;
  uid_t  ff_euid;

  SL_ENTER(_("tf_trust_check"));

  if (mode == SL_YESPRIV)
    sl_get_euid(&ff_euid);
  else
    sl_get_ruid(&ff_euid);

  status = sl_trustfile_euid(file, (long) ff_euid);

  if ( SL_ENONE != status) 
    {
      if (status == SL_ESTAT) 
	level = SH_ERR_ALL;
      else
	level = SH_ERR_ERR;

      tmp  = sh_util_safe_name (file);
      p    = sl_trust_errfile();
      if (p && *p != '\0')
	{
	  tmp2  = sh_util_safe_name (sl_trust_errfile());
	  sh_error_handle(level, FIL__, __LINE__, status, MSG_E_TRUST2,
			  sl_error_string(status), tmp, tmp2);
	  SH_FREE(tmp2);  
	}
      else
	{
	  sh_error_handle(level, FIL__, __LINE__, status, MSG_E_TRUST1,
			  sl_error_string(status), tmp);
	}
      SH_FREE(tmp);

      if (status == SL_EBADUID   || status == SL_EBADGID || 
	  status == SL_EBADOTH   || status == SL_ETRUNC  || 
	  status == SL_EINTERNAL )
	{
	  switch (status) {
	  case SL_EINTERNAL:
	    dlog(1, FIL__, __LINE__, 
		 _("An internal error occured in the trustfile function.\n"));
	    break;
	  case SL_ETRUNC:
	    tmp  = sh_util_safe_name (file);
	    dlog(1, FIL__, __LINE__, 
		 _("A filename truncation occured in the trustfile function.\nProbably the normalized filename for %s\nis too long. This may be due e.g. to deep or circular softlinks.\n"), 
		 tmp);
	    SH_FREE(tmp);
	    break;
	  case SL_EBADOTH:
	    tmp  = sh_util_safe_name (file);
	    p    = sl_trust_errfile();
	    dlog(1, FIL__, __LINE__, 
		 _("The path element: %s\nin the filename: %s is world writeable.\n"),
		 p, tmp);
	    SH_FREE(p);
	    SH_FREE(tmp);
	    break;
	  case SL_EBADUID:
	    tmp  = sh_util_safe_name (file);
	    p    = sl_trust_errfile();
	    dlog(1, FIL__, __LINE__, 
		 _("The owner (UID = %ld) of the path element: %s\nin the filename: %s\nis not in the list of trusted users.\nTo fix the problem, you can:\n - run ./configure again with the option --with-trusted=0,...,UID\n   where UID is the UID of the untrusted user, or\n - use the option TrustedUser=UID in the configuration file.\n"),
		 sl_trust_baduid(), p, tmp);
	    SH_FREE(p);
	    SH_FREE(tmp);
	  case SL_EBADGID:
	    tmp  = sh_util_safe_name (file);
	    p    = sl_trust_errfile();
	    dlog(1, FIL__, __LINE__, 
		 _("The path element: %s\nin the filename: %s\nis group writeable (GID = %ld), and at least one of the group\nmembers (UID = %ld) is not in the list of trusted users.\nTo fix the problem, you can:\n - run ./configure again with the option --with-trusted=0,...,UID\n   where UID is the UID of the untrusted user, or\n - use the option TrustedUser=UID in the configuration file.\n"),
		 p, tmp, sl_trust_badgid(), sl_trust_baduid());
	    SH_FREE(p);
	    SH_FREE(tmp);
	    break;
	  default:
	    break;
	  }
	    
	  SL_RETURN((-1), _("tf_trust_check"));
	}
    }

  SL_RETURN((0), _("tf_trust_check"));
}

void sh_unix_closeall (int fd)
{
  int fdx = fd;
#ifdef OPEN_MAX
  int fdlimit = OPEN_MAX;
#else
  int fdlimit = sysconf(_SC_OPEN_MAX);
#endif

  SL_ENTER(_("sh_unix_closeall"));

  /* can't happen - so fix it :-(
   */
  if (fdlimit < 0)
    fdlimit = 8;  /* POSIX lower limit */

  if (fdlimit > 65536)
    fdlimit = 65536;

  /* Close everything from fd (inclusive) up to fdlimit (exclusive). 
   */

  while (fd < fdlimit)
    close(fd++);

  sl_dropall (fdx);

  SL_RET0(_("sh_unix_closeall"));
}

static void sh_unix_setlimits(void)
{
  struct rlimit limits;

  SL_ENTER(_("sh_unix_setlimits"));

  limits.rlim_cur = RLIM_INFINITY;
  limits.rlim_max = RLIM_INFINITY;

#ifdef RLIMIT_CPU
  setrlimit (RLIMIT_CPU,     &limits);
#endif
#ifdef RLIMIT_FSIZE
  setrlimit (RLIMIT_FSIZE,   &limits);
#endif
#ifdef RLIMIT_DATA
  setrlimit (RLIMIT_DATA,    &limits);
#endif
#ifdef RLIMIT_STACK
  setrlimit (RLIMIT_STACK,   &limits);
#endif
#ifdef RLIMIT_RSS
  setrlimit (RLIMIT_RSS,     &limits);
#endif
#ifdef RLIMIT_NPROC
  setrlimit (RLIMIT_NPROC,   &limits);
#endif
#if defined(RLIMIT_NOFILE)
  setrlimit (RLIMIT_NOFILE,  &limits);
#elif defined(RLIMIT_OFILE)
  setrlimit (RLIMIT_OFILE,   &limits);
#endif
#ifdef RLIMIT_MEMLOCK
  setrlimit (RLIMIT_MEMLOCK, &limits);
#endif

#if !defined(SL_DEBUG)
  /* no core dumps
   */
  limits.rlim_cur = 0;
  limits.rlim_max = 0;
#ifdef RLIMIT_CORE
  setrlimit (RLIMIT_CORE,    &limits);
#endif
#else
#ifdef RLIMIT_CORE
  setrlimit (RLIMIT_CORE,    &limits);
#endif
#endif


  SL_RET0(_("sh_unix_setlimits"));
}

/* delete all environment variables
 */
static void sh_unix_zeroenv(void)
{
  char * c;
  char ** env = environ;

  SL_ENTER(_("sh_unix_zeroenv"));

  while (*env != NULL) {
    c = strchr ((*env), '=');
    ++c;
    ++(env);
    while ((*c) != '\0') {
      (*c) = '\0';
      ++c;
    }
  }
  SL_RET0(_("sh_unix_zeroenv"));
}


static void  sh_unix_resettimer(void)
{
  struct itimerval this_timer;

  SL_ENTER(_("sh_unix_resettimer"));

  this_timer.it_value.tv_sec  = 0;
  this_timer.it_value.tv_usec = 0;

  setitimer(ITIMER_REAL,    &this_timer, NULL);
#if !defined(SH_PROFILE)
  setitimer(ITIMER_VIRTUAL, &this_timer, NULL);
  setitimer(ITIMER_PROF,    &this_timer, NULL);
#endif

  SL_RET0(_("sh_unix_resettimer"));
}

static void  sh_unix_resetsignals(void)
{
  int  sig_num;
#ifdef NSIG
  int  max_sig = NSIG; 
#else
  int  max_sig = 255;
#endif
  int  test;
  struct sigaction act, oldact;
  int  status;

  sigset_t set_proc;

  SL_ENTER(_("sh_unix_resetsignals"));
  /* 
   * Reset the current signal mask (inherited from parent process).
   */

  sigfillset(&set_proc);

  do {
    errno = 0;
    test  = sigprocmask(SIG_UNBLOCK, &set_proc, NULL);
  } while (test < 0 && errno == EINTR);

  /* 
   * Reset signal handling.
   */
  
  act.sa_handler = SIG_DFL;         /* signal action           */
  sigemptyset( &act.sa_mask );      /* set an empty mask       */
  act.sa_flags = 0;                 /* init sa_flags           */

  for (sig_num = 1; sig_num <= max_sig; ++sig_num) 
    {
#if !defined(SH_PROFILE)
      test = retry_sigaction(FIL__, __LINE__, sig_num,  &act, &oldact);
#else
      test = 0;
#endif
      if ((test == -1) && (errno != EINVAL)) 
	{
	  status = errno;
	  sh_error_handle ((-1), FIL__, __LINE__, status, MSG_W_SIG,
			   sh_error_message (status), sig_num);
	}
    }

  SL_RET0(_("sh_unix_resetsignals"));
}

/* Get the local hostname (FQDN)
 */
#include <netdb.h>
#include <sys/socket.h> 

/* Required for BSD
 */
#ifdef HAVE_NETINET_IN_H
#include <netinet/in.h>
#endif

#include <arpa/inet.h>
#ifdef HAVE_UNAME
#include <sys/utsname.h>
void sh_unix_localhost()
{
  struct utsname   buf;
  struct hostent * he1;
  int              i;
  int              ddot = 0;
  int              len;


  SL_ENTER(_("sh_unix_localhost"));

  (void) uname (&buf);

  sl_strlcpy (sh.host.system,  buf.sysname, SH_MINIBUF);
  sl_strlcpy (sh.host.release, buf.release, SH_MINIBUF);
  sl_strlcpy (sh.host.machine, buf.machine, SH_MINIBUF);

  he1 = gethostbyname(buf.nodename);

  if (he1 == NULL)
    sl_strlcpy (sh.host.name, buf.nodename, SH_PATHBUF);
  else
    sl_strlcpy (sh.host.name, he1->h_name, SH_PATHBUF);
  

  /* check whether it looks like a FQDN
   */
  len = sl_strlen(sh.host.name);
  for (i = 0; i < len; ++i) 
    if (sh.host.name[i] == '.') ++ddot; 

  if (ddot == 0 && he1 != NULL) 
    sl_strlcpy (sh.host.name, 
	     inet_ntoa (*(struct in_addr *) he1->h_addr), 
	     SH_PATHBUF); 

  SL_RET0(_("sh_unix_localhost"));
}
#else
void sh_unix_localhost()
{
  struct hostent * he1;

  SL_ENTER(_("sh_unix_localhost"));

  (void) gethostname (sh.host.name, SH_PATHBUF-1);
  sh.host.name[SH_PATHBUF-1] = '\0';
  he1 = gethostbyname(sh.host.name);

  if (he1 != NULL)
    sl_strlcpy (sh.host.name, he1->h_name, SH_PATHBUF);
  else
    {
      sl_strlcpy (sh.host.name, _("localhost"), SH_PATHBUF);
      SL_RET0(_("sh_unix_localhost"));
    }

  /* check whether it looks like a FQDN
   */
  len = sl_strlen(sh.host.name);
  for (i = 0; i < len; ++i) 
    if (sh.host.name[i] == '.') ++ddot; 
  if (ddot == 0) 
    sl_strlcpy (sh.host.name, 
		inet_ntoa (*(struct in_addr *) he1->h_addr), 
		SH_PATHBUF); 

  SL_RET0(_("sh_unix_localhost"));
}
#endif


void sh_unix_mlock()
{
  SL_ENTER(_("sh_unix_mlock"));

#if defined(HAVE_MLOCK) && !defined(HAVE_BROKEN_MLOCK)
  if (skey->mlock_failed == GOOD)
    {
      sl_set_suid ();
      if ( (-1) == mlock( (char *) skey, sizeof (sh_key_t)) ) 
	{
	  skey->mlock_failed = BAD;
	}
      sl_unset_suid ();
    }
#else
  if (skey->mlock_failed == GOOD)
    {
      skey->mlock_failed = BAD;
    }
#endif

  SL_RET0(_("sh_unix_mlock"));
}
  
/* daemon mode 
 */
int sh_unix_setdeamon(char * dummy)
{
  static int block = 0;
  int    res = 0;

  SL_ENTER(_("sh_unix_setdeamon"));

  if (block != 0)
    SL_RETURN((0),_("sh_unix_setdeamon"));

  if (dummy == NULL)
    sh.flag.isdaemon = ON;
  else 
    res = sh_util_flagval (dummy, &sh.flag.isdaemon);
  /*
  if (0 == sl_strncmp(dummy, _("no"), 2))
    sh.flag.isdaemon = OFF;
  else
    sh.flag.isdaemon = ON;
  */

  if (sh.flag.opts == S_TRUE)  
    block = 1;
	   
  SL_RETURN(res, _("sh_unix_setdeamon"));
}

int sh_unix_init(int goDaemon)
{
  int    status;
  uid_t  uid;
  pid_t  oldpid = getpid();

  SL_ENTER(_("sh_unix_init"));

  /* fork twice, exit the parent process
   */
  if (goDaemon == 1) {
    
    switch (aud_fork(FIL__, __LINE__)) {
    case 0:  break;                             /* child process continues */
    case -1: SL_RETURN((-1),_("sh_unix_init")); /* error                   */
    default: aud__exit(FIL__, __LINE__, 0);     /* parent process exits    */
    }

    /* Child processes do not inherit page locks across a fork.
     * Error in next fork would return in this (?) thread of execution.
     */
    sh_unix_mlock();

    setsid();            /* should not fail         */

    switch (aud_fork(FIL__, __LINE__)) {
    case 0:  break;                             /* child process continues */
    case -1: SL_RETURN((-1),_("sh_unix_init")); /* error                   */
    default: aud__exit(FIL__, __LINE__, 0);     /* parent process exits    */
    }

    /* Child processes do not inherit page locks across a fork.
     */
    sh_unix_mlock();

  } else {
    setsid();            /* should not fail         */
  } 

  /* set working directory   
   */
  status = retry_aud_chdir(FIL__, __LINE__, "/");
  if ( (-1) == status ) 
    {
      status = errno;
      sh_error_handle ((-1), FIL__, __LINE__, status, MSG_W_CHDIR,
		       sh_error_message (status), "/");
      aud_exit(FIL__, __LINE__, EXIT_FAILURE);
    }

  /* reset timers 
   */
  sh_unix_resettimer();

  /* signal handlers 
   */
  sh_unix_resetsignals();
  sh_unix_siginstall  ();

  /* set file creation mask 
   */
  (void) umask (0); /* should not fail */

  /* set resource limits to maximum, and
   * core dump size to zero 
   */
  sh_unix_setlimits();

  /* zero out the environment (like PATH='\0')  
   */
  sh_unix_zeroenv();


  /* close all file descriptors, and 
   * open first three streams to /dev/null 
   */
  if (goDaemon == 1)
    {
      sh_unix_closeall (0);
      aud_open(FIL__, __LINE__, SL_NOPRIV, _("/dev/null"), O_RDWR, 0);
      retry_aud_dup(FIL__, __LINE__, 0); 
      retry_aud_dup(FIL__, __LINE__, 0);

      /* --- wait until parent has exited ---
       */
      while (1 == 1)
	{
	  errno = 0;
	  if (0 > aud_kill (FIL__, __LINE__, oldpid, 0) && errno == ESRCH)
	    {
	      break;
	    }
#ifdef HAVE_USLEEP
	  usleep(100);
#else
	  sleep(1);
#endif
	}

      /* write PID file
       */
      status = sh_unix_testlock(NULL);
      if (status < 0)
	{
	  sl_get_euid(&uid);
	  sh_error_handle ((-1), FIL__, __LINE__, status, MSG_PIDFILE,
			   (long) uid, sh.srvlog.alt);
	  aud_exit(FIL__, __LINE__, EXIT_FAILURE);
	}
    }

  SL_RETURN((0),_("sh_unix_init"));
}

/********************************************************
 *
 *  TIME
 *
 ********************************************************/

/* Figure out the time offset of the current timezone
 * in a portable way.
 */
char * t_zone(const time_t * xx)
{
  struct tm   aa;
  struct tm   bb;
  struct tm * cc;
  int  sign =  0;
  int  diff =  0;
  int  hh, mm;
  static char tz[64];

  SL_ENTER(_("t_zone"));


  cc = gmtime (xx);
  memcpy (&aa, cc, sizeof(struct tm));
  cc = localtime (xx);
  memcpy (&bb, cc, sizeof(struct tm));

  /* Check for datum wrap-around.
   */
  if      (aa.tm_year < bb.tm_year)
    sign = (-1);
  else if (aa.tm_mon  < bb.tm_mon)
    sign = (-1);
  else if (aa.tm_mday < bb.tm_mday)
    sign = (-1);
  else if (bb.tm_year < aa.tm_year)
    sign = ( 1);
  else if (bb.tm_mon  < aa.tm_mon)
    sign = ( 1);
  else if (bb.tm_mday < aa.tm_mday)
    sign = ( 1);

  diff = aa.tm_hour * 60 + aa.tm_min;
  diff = (bb.tm_hour * 60 + bb.tm_min) - diff;
  diff = diff - (sign * 24 * 60);   /* datum wrap-around correction */
  hh = diff / 60;
  mm = diff - (hh * 60);
  sprintf (tz, _("%+03d%02d"), hh, mm);                /* known to fit  */

  SL_RETURN(tz, _("t_zone"));
}

unsigned long sh_unix_longtime ()
{
  return ((unsigned long)time(NULL));
} 

#ifdef HAVE_GETTIMEOFDAY
unsigned long sh_unix_notime ()
{
  struct timeval  tv;
  struct timezone tz;

  tz.tz_minuteswest = 0;
  tz.tz_dsttime     = 0;

  gettimeofday (&tv, &tz);

  return ((unsigned long)(tv.tv_sec + tv.tv_usec * 10835 + getpid() + getppid()));
  
}
#endif

static int count_dev_time = 0;

void reset_count_dev_time(void)
{
  count_dev_time = 0;
  return;
}

int sh_unix_settimeserver (char * address)
{

  SL_ENTER(_("sh_unix_settimeserver"));

  if (address != NULL && count_dev_time < 2 
      && sl_strlen(address) < SH_PATHBUF) 
    {
      if (count_dev_time == 0)
	sl_strlcpy (sh.srvtime.name, address, SH_PATHBUF);
      else
	sl_strlcpy (sh.srvtime.alt,  address, SH_PATHBUF);

      ++count_dev_time;
      SL_RETURN((0), _("sh_unix_settimeserver"));
    }
  SL_RETURN((-1), _("sh_unix_settimeserver"));
}


#ifdef HAVE_NTIME
#define UNIXEPOCH 2208988800UL  /* difference between Unix time and net time 
                                 * The UNIX EPOCH starts in 1970.
                                 */
#include <sys/socket.h>
#include <netinet/in.h>
#include <arpa/inet.h>
#include <netdb.h>
#include <ctype.h>
#endif

/* Timeserver service.               */
/* define is missing on HP-UX 10.20  */
#ifndef IPPORT_TIMESERVER 
#define IPPORT_TIMESERVER 37 
#endif

char * sh_unix_time (time_t thetime)
{

  int           status;

  time_t        time_now;
  struct tm   * time_ptr;
  static char   AsciiTime[81];                       /* local time   */
  static char   RetTime[81];                         /* local time   */
#ifdef SH_USE_XML
  static char   deftime[] = N_("0000-00-00T00:00:00"); /* default time */
#else
  static char   deftime[] = N_("[0000-00-00T00:00:00]"); /* default time */
#endif

#ifdef HAVE_NTIME
  int    fd;                    /* network file descriptor                  */
  u_char net_time[4];           /* remote time in network format            */
  static int failerr = 0;       /* no net time                              */
  int    fail = 0;              /* no net time                              */
  int    errflag;
  char   errmsg[256];
  char   error_call[SH_MINIBUF];
  int    error_num;
#endif
  
  SL_ENTER(_("sh_unix_time"));

#ifdef HAVE_NTIME
  if (thetime == 0) 
    {
      if (sh.srvtime.name[0] == '\0') 
	{
	  fail = 1;
	  (void) time (&time_now);
	} 
      else /* have a timeserver address */
	{ 
	  fd = connect_port_2 (sh.srvtime.name, sh.srvtime.alt, 
			       IPPORT_TIMESERVER, 
			       error_call, &error_num, errmsg, sizeof(errmsg));
	  if (fd >= 0)
	    {
	      if (4 != read_port (fd, (char *) net_time, 4, &errflag, 2))
		{
		  fail = 1;
		  sh_error_handle ((-1), FIL__, __LINE__, errflag, 
				   MSG_E_NLOST, 
				   _("time"), sh.srvtime.name);
		}
	      close(fd);
	    }
	  else
	    {
	      sh_error_handle ((-1), FIL__, __LINE__, error_num, 
			       MSG_E_NET, errmsg, error_call,
			       _("time"), sh.srvtime.name);
	      fail = 1;
	    }
	  
	  if (fail == 0) 
	    { 
	      time_now = ntohl(* (long *) net_time) - UNIXEPOCH;
	      /* fprintf(stderr, "TIME IS %ld\n", time_now); */
	      if (failerr == 1) {
		failerr = 0;
		sh_error_handle ((-1), FIL__, __LINE__, 0, 
				 MSG_E_NEST, 
				 _("time"), sh.srvtime.name);
	      } 
	    }
	  else
	    {
	      (void) time (&time_now);
	      if (failerr == 0)
		{
		  failerr = 1;
		  sh_error_handle ((-1), FIL__, __LINE__, errflag, 
				   MSG_SRV_FAIL, 
				   _("time"), sh.srvtime.name);
		}
	    }
	}
    }
  else 
    {
      time_now = thetime;
    }

  /* #ifdef HAVE_NTIME */
#else

  if (thetime == 0) 
    {
      (void) time (&time_now);
    } 
  else 
    {
      time_now = thetime;
    }

  /* #ifdef HAVE_NTIME */
#endif

  if (time_now == (-1) ) 
    SL_RETURN( _(deftime), _("sh_unix_time"));
  else
    time_ptr   = localtime (&time_now);

  if (time_ptr != NULL) 
    {
      status = strftime (AsciiTime, 80,
#ifdef SH_USE_XML
			 _("%Y-%m-%dT%H:%M:%S%%s"),
#else
			 _("[%Y-%m-%dT%H:%M:%S%%s]"),
#endif
			 time_ptr);

      sl_snprintf(RetTime, 80, AsciiTime, t_zone(&time_now));

      if ( (status == 0) || (status == 80) ) 
	SL_RETURN( _(deftime), _("sh_unix_time"));
      else
	SL_RETURN( &RetTime[0], _("sh_unix_time"));
    }

  /* last resort
   */
  SL_RETURN( _(deftime), _("sh_unix_time"));
}

static int sh_unix_use_localtime = S_FALSE;

/* whether to use localtime for file timesatams in logs
 */
int sh_unix_uselocaltime (char * c)
{
  int i;
  SL_ENTER(_("sh_unix_uselocaltime"));
  i = sh_util_flagval(c, &(sh_unix_use_localtime));

  SL_RETURN(i, _("sh_unix_uselocaltime"));
}
    
char * sh_unix_gmttime (time_t thetime)
{

  int           status;

  struct tm   * time_ptr;
  static char   AsciiTime[81];                       /* GMT time   */
#ifdef SH_USE_XML
  static char   deftime[] = N_("0000-00-00T00:00:00"); /* default time */
#else
  static char   deftime[] = N_("[0000-00-00T00:00:00]"); /* default time */
#endif

  SL_ENTER(_("sh_unix_gmttime"));

  if (sh_unix_use_localtime == S_FALSE)
    time_ptr   = gmtime (&thetime);
  else
    time_ptr   = localtime (&thetime);

  if (time_ptr != NULL) 
    {
      status = strftime (AsciiTime, 80,
#ifdef SH_USE_XML
			 _("%Y-%m-%dT%H:%M:%S"),
#else
			 _("[%Y-%m-%dT%H:%M:%S]"),
#endif
			 time_ptr);

      if ( (status == 0) || (status == 80) ) 
	SL_RETURN( _(deftime), _("sh_unix_gmttime"));
      else
	SL_RETURN( &AsciiTime[0], _("sh_unix_gmttime"));
    }

  /* last resort
   */
  SL_RETURN( _(deftime), _("sh_unix_gmttime"));
}



char *  sh_unix_getUIDdir (int level, uid_t uid)
{
  struct passwd * tempres;

  SL_ENTER(_("sh_unix_getUIDdir"));

  tempres = getpwuid(uid);

  if (tempres == NULL) {
    sh_error_handle (level, FIL__, __LINE__, EINVAL, MSG_E_PWNULL,
		     _("getpwuid"), (long) uid, _("completely missing"));
    SL_RETURN( NULL, _("sh_unix_getUIDdir"));
  }

  if (tempres->pw_dir != NULL) {
    SL_RETURN( tempres->pw_dir, _("sh_unix_getUIDdir"));
  } else {
    sh_error_handle (level, FIL__, __LINE__, EINVAL, MSG_E_PWNULL,
		     _("getpwuid"), (long) uid, _("pw_dir"));
    SL_RETURN( NULL, _("sh_unix_getUIDdir"));
  }
}

char *  sh_unix_getUIDname (int level, uid_t uid)
{
  struct passwd * tempres;

  SL_ENTER(_("sh_unix_getUIDname"));

  tempres = getpwuid(uid);

  if (tempres == NULL) {
    sh_error_handle (level, FIL__, __LINE__, EINVAL, MSG_E_PWNULL,
		     _("getpwuid"), (long) uid, _("completely missing"));
    SL_RETURN( NULL, _("sh_unix_getUIDname"));
  }

  if (tempres->pw_name != NULL) {
    SL_RETURN( tempres->pw_name, _("sh_unix_getUIDname"));
  } else {
    sh_error_handle (level, FIL__, __LINE__, EINVAL, MSG_E_PWNULL,
		     _("getpwuid"), (long) uid, _("pw_user"));
    SL_RETURN( NULL, _("sh_unix_getUIDname"));
  }
}

char *  sh_unix_getGIDname (int level, gid_t gid)
{
  struct group * tempres;

  SL_ENTER(_("sh_unix_getGIDname"));

  tempres = getgrgid(gid);

  if (tempres == NULL) {
    sh_error_handle (level, FIL__, __LINE__, EINVAL, MSG_E_GRNULL,
		     _("getgrgid"), (long) gid, _("completely missing"));
    SL_RETURN( NULL, _("sh_unix_getGIDname"));
  }

  if (tempres->gr_name != NULL) {
    SL_RETURN( tempres->gr_name, _("sh_unix_getGIDname"));
  } else {
    sh_error_handle (level, FIL__, __LINE__, EINVAL, MSG_E_GRNULL,
		     _("getgrgid"), (long) gid, _("gr_name"));
    SL_RETURN( NULL, _("sh_unix_getGIDname"));
  }
}

int sh_unix_getUser ()
{
  char          * p;
  uid_t  seuid, sruid;

  SL_ENTER(_("sh_unix_getUser"));

  (void) sl_get_euid(&seuid);

  sh.effective.uid = seuid;

  p = sh_unix_getUIDdir (SH_ERR_ERR, seuid);

  if (p == NULL)
    SL_RETURN((-1), _("sh_unix_getUser"));
  else
    {
      if (sl_strlen(p) >= SH_PATHBUF) {
	sh_error_handle (SH_ERR_ERR, FIL__, __LINE__, EINVAL, MSG_E_PWLONG,
			 _("getpwuid"), (long) seuid, _("pw_home"));
	SL_RETURN((-1), _("sh_unix_getUser"));
      } else {
	sl_strlcpy ( sh.effective.home, p, SH_PATHBUF);
      }
    }

  (void) sl_get_ruid(&sruid);

  sh.real.uid = sruid;

  p = sh_unix_getUIDname (SH_ERR_ERR, sruid);
  if (p == NULL)
    SL_RETURN((-1), _("sh_unix_getUser"));
  else
    {
      if (sl_strlen(p) >= USER_MAX) {
	sh_error_handle (SH_ERR_ERR, FIL__, __LINE__, EINVAL, MSG_E_PWLONG,
			 _("getpwuid"), (long) sruid, _("pw_user"));
	SL_RETURN((-1), _("sh_unix_getUser"));
      } else {
	sl_strlcpy ( sh.real.user, p, USER_MAX);
      }
    }

  p = sh_unix_getUIDdir (SH_ERR_ERR, sruid);

  if (p == NULL)
    SL_RETURN((-1), _("sh_unix_getUser"));
  else
    {
      if (sl_strlen(p) >= SH_PATHBUF) {
	sh_error_handle (SH_ERR_ERR, FIL__, __LINE__, EINVAL, MSG_E_PWLONG,
			 _("getpwuid"), (long) sruid, _("pw_home"));
	SL_RETURN((-1), _("sh_unix_getUser"));
      } else {
	sl_strlcpy ( sh.real.home, p, SH_PATHBUF);
      }
    }

  SL_RETURN((0), _("sh_unix_getUser"));

  /* notreached */
}



int sh_unix_getline (SL_TICKET fd, char * line, int sizeofline)
{
  register int  count;
  register int  n = 0;
  char          c;

  SL_ENTER(_("sh_unix_getline"));

  if (sizeofline < 2) {
    line[n] = '\0';
    SL_RETURN((0), _("sh_unix_getline"));
  }

  while (n < sizeofline) {

    count = sl_read (fd, &c, 1);

    /* fprintf (stderr, "%c", c); */

    /* end of file
     */
    if (count < 1) {
      line[n] = '\0';
      n = -1;
      break;
    } 

    if (/* c != '\0' && */ c != '\n') {
      line[n] = c;
      ++n;
    } else if (c == '\n') {
      if (n > 0) {
	line[n] = '\0';
	break;
      } else {
	line[n] = '\n'; /* get newline only if only char on line */
	++n;
	line[n] = '\0';
	break;
      }
    } else {
      line[n] = '\0';
      break;
    }

  }


  line[sizeofline] = '\0';  /* make sure line is terminated */
  SL_RETURN((n), _("sh_unix_getline"));
}


#if defined (SH_WITH_CLIENT) || defined (SH_STANDALONE) 

/**************************************************************
 *
 * --- FILE INFO ---
 *
 **************************************************************/

#if defined(__linux__)

/* --- Determine ext2fs file attributes. ---
 */
#include <sys/ioctl.h>
#include <linux/ext2_fs.h>

static 
int sh_unix_getinfo_attr (char * name, 
			  unsigned long * flags, 
			  char * c_attr)
{

/* TAKEN FROM:
 *
 * lsattr.c             - List file attributes on an ext2 file system
 *
 * Copyright (C) 1993, 1994  Remy Card <card@masi.ibp.fr>
 *                           Laboratoire MASI, Institut Blaise Pascal
 *                           Universite Pierre et Marie Curie (Paris VI)
 *
 * This file can be redistributed under the terms of the GNU General
 * Public License
 */

#ifdef HAVE_STAT_FLAGS
  struct stat buf;
  int    i;

  SL_ENTER(_("sh_unix_getinfo_attr"));

  *flags = 0;

  sl_set_suid();                   /* sl_set/unset_suid inserted R.Wichmann */
  i = retry_stat (FIL__, __LINE__, name, &buf);
  sl_unset_suid();

  if (i == -1)
    SL_RETURN(-1, _("sh_unix_getinfo_attr"));

#ifdef UF_IMMUTABLE
  if (buf.st_flags & UF_IMMUTABLE)
    *flags |= EXT2_IMMUTABLE_FL;
#endif
#ifdef UF_APPEND
  if (buf.st_flags & UF_APPEND)
    *flags |= EXT2_APPEND_FL;
#endif
#ifdef UF_NODUMP
  if (buf.st_flags & UF_NODUMP)
    *flags |= EXT2_NODUMP_FL;
#endif

#else

#ifdef HAVE_EXT2_IOCTLS
  int fd, r, f;
  
  SL_ENTER(_("sh_unix_getinfo_attr"));

  *flags = 0;

  /* open() -> aud_open() R.Wichmann */
  fd = aud_open (FIL__, __LINE__, SL_YESPRIV, name, O_RDONLY|O_NONBLOCK, 0);

  if (fd == -1)
    SL_RETURN(-1, _("sh_unix_getinfo_attr"));

  r = ioctl (fd, EXT2_IOC_GETFLAGS, &f);
  close (fd);

  if (r == -1)
    SL_RETURN(-1, _("sh_unix_getinfo_attr"));

  *flags = f;

/* ! HAVE_EXT2_IOCTLS */
#else 

  SL_ENTER(_("sh_unix_getinfo_attr"));

  *flags = 0;                                     /* modified by R.Wichmann */

/* ! HAVE_EXT2_IOCTLS */
#endif 
#endif
/*
 * END
 *
 * lsattr.c             - List file attributes on an ext2 file system
 */

#ifdef EXT2_SECRM_FL
  if ( (*flags & EXT2_SECRM_FL) != 0  )   c_attr[0] = 's';
#endif
#ifdef EXT2_UNRM_FL 
  if ( (*flags & EXT2_UNRM_FL) != 0   )   c_attr[1] = 'u';
#endif
#ifdef EXT2_SYNC_FL
  if ( (*flags & EXT2_SYNC_FL) != 0    )  c_attr[2] = 'S';
#endif
#ifdef EXT2_IMMUTABLE_FL
  if ( (*flags & EXT2_IMMUTABLE_FL) != 0) c_attr[3] = 'i';
#endif
#ifdef EXT2_APPEND_FL
  if ( (*flags & EXT2_APPEND_FL) != 0  )  c_attr[4] = 'a';
#endif
#ifdef EXT2_NODUMP_FL
  if ( (*flags & EXT2_NODUMP_FL) != 0  )  c_attr[5] = 'd';
#endif
#ifdef EXT2_NOATIME_FL
  if ( (*flags & EXT2_NOATIME_FL) != 0)   c_attr[6] = 'A';
#endif
#ifdef EXT2_COMPR_FL
  if ( (*flags & EXT2_COMPR_FL) != 0   )  c_attr[7] = 'c';
#endif
#ifdef EXT2_COMPRBLK_FL
  if ( (*flags & EXT2_COMPRBLK_FL) != 0)  c_attr[8] = 'B';
#endif
#ifdef EXT2_DIRTY_FL
  if ( (*flags & EXT2_DIRTY_FL) != 0   )  c_attr[9] = 'D';
#endif
#ifdef EXT2_NOCOMPR_FL
  if ( (*flags & EXT2_NOCOMPR_FL) != 0 )  c_attr[10] = 'D';
#endif
#ifdef EXT2_ECOMPR_FL
  if ( (*flags & EXT2_ECOMPR_FL) != 0  )  c_attr[11] = 'D';
#endif

  c_attr[12] = '\0';

  SL_RETURN(0, _("sh_unix_getinfo_attr"));
}
#endif

/* determine file type
 */
static 
int sh_unix_getinfo_type (struct stat * buf, 
			  ShFileType * type, 
			  char * c_mode)
{
  SL_ENTER(_("sh_unix_getinfo_type"));

  if      ( S_ISREG(buf->st_mode)  ) { 
    (*type)   = SH_FILE_REGULAR;
    c_mode[0] = '-';
  }
  else if ( S_ISLNK(buf->st_mode)  ) {
    (*type)   = SH_FILE_SYMLINK;
    c_mode[0] = 'l';
  }
  else if ( S_ISDIR(buf->st_mode)  ) {
    (*type)   = SH_FILE_DIRECTORY;
    c_mode[0] = 'd';
  }
  else if ( S_ISCHR(buf->st_mode)  ) {
    (*type)   = SH_FILE_CDEV;
    c_mode[0] = 'c';
  }
  else if ( S_ISBLK(buf->st_mode)  ) {
    (*type)   = SH_FILE_BDEV;
    c_mode[0] = 'b';
  }
  else if ( S_ISFIFO(buf->st_mode) ) {
    (*type)   = SH_FILE_FIFO;
    c_mode[0] = '|';
  }
  else if ( S_ISSOCK(buf->st_mode) ) {
    (*type)   = SH_FILE_SOCKET;
    c_mode[0] = 's';
  }
  else                              {
    (*type)   = SH_FILE_UNKNOWN;
    c_mode[0] = '?';
  }

  SL_RETURN(0, _("sh_unix_getinfo_type"));
}

int sh_unix_get_ftype(char * fullpath)
{
  char        c_mode[16];
  struct stat buf;
  ShFileType  type;
  int         res;

  SL_ENTER(_("sh_unix_get_ftype"));

  sl_set_suid();
  res = retry_lstat(FIL__, __LINE__, fullpath, &buf);
  sl_unset_suid();

  if (res < 0)
    SL_RETURN(SH_FILE_UNKNOWN, _("sh_unix_getinfo_type"));

  sh_unix_getinfo_type (&buf, &type, c_mode);

  SL_RETURN(type, _("sh_unix_get_ftype"));
}


static 
int  sh_unix_getinfo_mode (struct stat *buf, 
			   unsigned int * mode, 
			   char * c_mode)
{

  SL_ENTER(_("sh_unix_getinfo_mode"));

  (*mode) = buf->st_mode;

  /* make 'ls'-like string */
  
  if ( (buf->st_mode & S_IRUSR) != 0 )  c_mode[1] = 'r'; 
  if ( (buf->st_mode & S_IWUSR) != 0 )  c_mode[2] = 'w'; 
  if ( (buf->st_mode & S_IXUSR) != 0 ) {
    if ((buf->st_mode & S_ISUID) != 0 ) c_mode[3] = 's';
    else                                c_mode[3] = 'x';
  } else {
    if ((buf->st_mode & S_ISUID) != 0 ) c_mode[3] = 'S';
  }

  if ( (buf->st_mode & S_IRGRP) != 0 )  c_mode[4] = 'r'; 
  if ( (buf->st_mode & S_IWGRP) != 0 )  c_mode[5] = 'w'; 
  if ( (buf->st_mode & S_IXGRP) != 0 )  {
    if ((buf->st_mode & S_ISGID) != 0 ) c_mode[6] = 's';
    else                                c_mode[6] = 'x';
  } else {
    if ((buf->st_mode & S_ISGID) != 0 ) c_mode[6] = 'S';
  } 

  if ( (buf->st_mode & S_IROTH) != 0 )  c_mode[7] = 'r'; 
  if ( (buf->st_mode & S_IWOTH) != 0 )  c_mode[8] = 'w';
#ifdef S_ISVTX  /* not POSIX */
  if ( (buf->st_mode & S_IXOTH) != 0 )  {
    if ((buf->st_mode & S_ISVTX) != 0 ) c_mode[9] = 't';
    else                                c_mode[9] = 'x';
  } else {
    if ((buf->st_mode & S_ISVTX) != 0 ) c_mode[9] = 'T';
  }
#else
  if ( (buf->st_mode & S_IXOTH) != 0 )  c_mode[9] = 'x';
#endif 

  SL_RETURN(0, _("sh_unix_getinfo_mode"));
}

static long IO_Limit = 0;

int sh_unix_set_io_limit (char * c)
{
  long val;

  SL_ENTER(_("sh_unix_set_io_limit"));

  val = strtol (c, (char **)NULL, 10);
  if (val < 0)
    sh_error_handle ((-1), FIL__, __LINE__, EINVAL, MSG_EINVALS,
                      _("set I/O limit"), c);

  val = (val < 0 ? 0 : val);

  IO_Limit = val * 1000;
  SL_RETURN( 0, _("sh_unix_set_io_limit"));
}

#include <setjmp.h>

static jmp_buf proc_wait_timeout;

static void sh_unix_proc_alarmhandle (int mysignal)
{
  struct  sigaction  def_act;
  def_act.sa_handler = SIG_IGN;

  if (mysignal == SIGALRM)   /* use 'signal' to avoid compiler warning */
    {  
      alarm(0);
      sigaction (SIGALRM, &def_act, NULL);
    }
  longjmp(proc_wait_timeout, 1);
}


/* obtain file info
 */
int sh_unix_getinfo (int level, 
		     char * filename, file_type * theFile, char * fileHash)
{
  long          runtim;
  struct stat   buf;
  struct stat   lbuf;
  struct stat   fbuf;
  int           stat_return;
  register int  i;
  ShFileType    type;
  unsigned int  mode;
  char        * name;
  char        * tmp;
  char        * tmp2;
  char        * p;
  char        * linknamebuf;
  int           linksize;

  int get_the_fd (SL_TICKET ticket);

  SL_TICKET     rval_open;
  SL_TICKET     rval_open_2;
  int           fd;
  int           fstat_return;

  time_t        tend;
  time_t        tstart;

  struct  sigaction  def_act;
  struct  sigaction  new_act;
  sigset_t           unblock;

  def_act.sa_handler = SIG_IGN;
  new_act.sa_handler = sh_unix_proc_alarmhandle;

  sigemptyset( &new_act.sa_mask );       /* set an empty mask       */
  new_act.sa_flags = 0;                  /* init sa_flags           */

  sigemptyset(&unblock);
  sigaddset  (&unblock, SIGALRM);

  SL_ENTER(_("sh_unix_getinfo"));

  /* --- Stat the file, and get checksum. ---
   */
  tstart = time(NULL);
  sl_set_suid();
  stat_return = retry_lstat (FIL__, __LINE__, theFile->fullpath, &buf);
  sl_unset_suid();

  if (stat_return == 0 && S_ISREG(buf.st_mode)) {
    rval_open = sl_open_fastread (theFile->fullpath, SL_YESPRIV);
  }
  else
    rval_open = -1;

  rval_open_2 = rval_open;

  if (SL_ISERROR(rval_open))
    fd = -1;
  else
    fd = get_the_fd(rval_open);

  tend = time(NULL);

  /* An unprivileged user may slow lstat/open to a crawl
   * with clever path/symlink setup
   */
  if ((tend - tstart) > (time_t) 60)
    {
      tmp2 = sh_util_safe_name (theFile->fullpath);
      sh_error_handle ((-1), FIL__, __LINE__, 0, MSG_FI_TOOLATE,
		       (long)(tend - tstart), tmp2);
      SH_FREE(tmp2);
    }
      
  if (fd >= 0) 
    {
      sl_set_suid();
      fstat_return = retry_fstat (FIL__, __LINE__, fd, &fbuf);
      sl_unset_suid();
    }
  else
    fstat_return = -1;


  /* ---  case 1: lstat failed  --- 
   */
  if (stat_return != 0) 
    {
      stat_return = errno;
      if (!SL_ISERROR(rval_open))
	  sl_close(rval_open);
      if (sh.flag.checkSum == SH_CHECK_INIT || 
	  (sh_hash_have_it (theFile->fullpath) >= 0 && 
	   theFile->reported == S_FALSE))
	{
	  tmp2 = sh_util_safe_name (theFile->fullpath);
	  sh_error_handle (level, FIL__, __LINE__, stat_return, MSG_FI_LSTAT,
			   sh_error_message (stat_return), tmp2);
	  SH_FREE(tmp2);
	}
      SL_RETURN((-1),_("sh_unix_getinfo"));
    }

  /* ---  case 2: not a regular file  --- 
   */
  else if (! S_ISREG(buf.st_mode) )
    {
      if (fileHash != NULL)
	sl_strlcpy(fileHash, 
		   _("000000000000000000000000000000000000000000000000"), 
		   KEY_LEN + 1);
    }
  
  /* ---  case 3a: a regular file, fstat ok --- 
   */
  else if (fstat_return == 0 && 
	   buf.st_mode == fbuf.st_mode &&
	   buf.st_ino  == fbuf.st_ino  &&
	   buf.st_uid  == fbuf.st_uid  &&
	   buf.st_gid  == fbuf.st_gid  &&
	   buf.st_dev  == fbuf.st_dev )
    {
      if (fileHash != NULL)
	{
	  tiger_fd = rval_open;
	  p = theFile->fullpath;

	  if (setjmp(proc_wait_timeout) != 0) {
	    TPT((0, FIL__, __LINE__, _("msg=<proc_wait: timeout>\n"))); 
	    sl_strlcpy(fileHash, 
		       _("000000000000000000000000000000000000000000000000"),
		       KEY_LEN + 1);
	    alarm(0);
	    sigaction (SIGALRM, &def_act, NULL);
	    sigprocmask(SIG_UNBLOCK, &unblock, NULL);
	    goto time_out_door;
	  }
	  alarm(0);
	  sigaction (SIGALRM, &new_act, &def_act);
	  alarm(120);

	  sl_strlcpy(fileHash, 
		     sh_tiger_generic_hash (theFile->fullpath, TIGER_FD, 0), 
		     KEY_LEN + 1);
	  p = theFile->fullpath;

	  alarm(0);
	  sigaction (SIGALRM, &def_act, NULL);

	}
    }

  /* ---  case 3b: a regular file, fstat ok, but different --- 
   */
  else if (fstat_return == 0 && S_ISREG(fbuf.st_mode))
    {
      memcpy (&buf, &fbuf, sizeof( struct stat ));
      p = theFile->fullpath;
      if (fileHash != NULL)
	{
	  tiger_fd = rval_open;

	  if (setjmp(proc_wait_timeout) != 0) {
	    TPT((0, FIL__, __LINE__, _("msg=<proc_wait: timeout>\n"))); 
	    sl_strlcpy(fileHash, 
		       _("000000000000000000000000000000000000000000000000"),
		       KEY_LEN + 1);
	    alarm(0);
	    sigaction (SIGALRM, &def_act, NULL);
	    goto time_out_door;
	  }
	  alarm(0);
	  sigaction (SIGALRM, &new_act, &def_act);
	  alarm(120);
	  
	  sl_strlcpy(fileHash, 
		     sh_tiger_generic_hash (theFile->fullpath, TIGER_FD, 0), 
		     KEY_LEN + 1);
	  p = theFile->fullpath;

	  alarm(0);
	  sigaction (SIGALRM, &def_act, NULL);
	  
	}
    }

  /* ---  case 4: a regular file, fstat failed --- 
   */

  else    /* fstat_return != 0 or !S_ISREG(fbuf->st_mode) */
    {
      fstat_return = errno;
      if (fileHash != NULL)
	sl_strlcpy(fileHash, 
		   _("000000000000000000000000000000000000000000000000"), 
		   KEY_LEN + 1);

      tmp2 = sh_util_safe_name (theFile->fullpath);
      sh_error_handle (level, FIL__, __LINE__, fstat_return, MSG_E_READ,
		       tmp2);
      SH_FREE(tmp2);
    }	  

 time_out_door:
  rval_open = rval_open_2;

  if (!SL_ISERROR(rval_open))
    sl_close(rval_open);


  /* --- I/O limit. --- 
   */
  if (IO_Limit > 0)
    {
      runtim = (long) (time(NULL) - sh.stat.time_start);
      
      if (runtim > 0 && (long)(sh.stat.bytes_hashed/runtim) > IO_Limit)
	sleep(1);
    }

  for (i = 0; i < 10; ++i) theFile->c_mode[i] = '-';
  theFile->c_mode[10] = '\0';
  for (i = 0; i < 11; ++i) theFile->link_c_mode[i] = '-';
  theFile->link_c_mode[10] = '\0';


  /* --- Determine file type. ---
   */
  sh_unix_getinfo_type (&buf, &type, theFile->c_mode);
  theFile->type = type;

#if defined(__linux__)

  /* --- Determine file attributes. ---
   */
  for (i = 0; i < 12; ++i) theFile->c_attributes[i] = '-';
  theFile->c_attributes[12] = '\0';
  theFile->attributes      =    0;
  if (theFile->c_mode[0] != 'c' && theFile->c_mode[0] != 'b' &&
      theFile->c_mode[0] != 'l' )
    sh_unix_getinfo_attr(theFile->fullpath, 
			 &theFile->attributes, theFile->c_attributes);
#endif


  /* --- Determine permissions. ---
   */
  sh_unix_getinfo_mode (&buf, &mode, theFile->c_mode);

  /* --- Trivia. ---
   */
  theFile->dev       = buf.st_dev;
  theFile->ino       = buf.st_ino;
  theFile->mode      = buf.st_mode;
  theFile->hardlinks = buf.st_nlink;
  theFile->owner     = buf.st_uid;  
  theFile->group     = buf.st_gid;  
  theFile->rdev      = buf.st_rdev;
  theFile->size      = buf.st_size;
  theFile->blksize   = (unsigned long) buf.st_blksize;
  theFile->blocks    = (unsigned long) buf.st_blocks;
  theFile->atime     = buf.st_atime;
  theFile->mtime     = buf.st_mtime;
  theFile->ctime     = buf.st_ctime;


  /* --- Owner and group. ---
   */

  if ( (name = sh_unix_getGIDname(ShDFLevel[SH_ERR_T_NAME], buf.st_gid)) != NULL) {
    sl_strlcpy (theFile->c_group, name, GROUP_MAX+1);
  } else {
    tmp2 = sh_util_safe_name (theFile->fullpath);
    /* level -> ShDFLevel[SH_ERR_T_NAME] */
    sh_error_handle (ShDFLevel[SH_ERR_T_NAME], FIL__, __LINE__, ENOENT, 
		     MSG_FI_NOGRP,
		     (long) buf.st_gid, tmp2);
    SH_FREE(tmp2);
    sl_snprintf(theFile->c_group, GROUP_MAX+1, "%d", (long) buf.st_gid); 
    /* theFile->c_group[0] = '\0'; */
  }

  
  if ( (name = sh_unix_getUIDname(ShDFLevel[SH_ERR_T_NAME], buf.st_uid)) != NULL) {
    sl_strlcpy (theFile->c_owner, name, USER_MAX+1);
  } else {
    tmp2 = sh_util_safe_name (theFile->fullpath);
    /* level -> ShDFLevel[SH_ERR_T_NAME] */
    sh_error_handle (ShDFLevel[SH_ERR_T_NAME], FIL__, __LINE__, ENOENT, 
		     MSG_FI_NOUSR,
		     (long) buf.st_uid, tmp2);
    SH_FREE(tmp2);
    sl_snprintf(theFile->c_owner, USER_MAX+1, "%d", (long) buf.st_uid); 
    /* theFile->c_owner[0] = '\0'; */
  }

  /* --- Output the file. ---
   */
  tmp2 = sh_util_safe_name (filename);
  sh_error_handle ((-1), FIL__, __LINE__, 0, MSG_FI_LIST,
		   theFile->c_mode,
		   theFile->hardlinks,
		   theFile->c_owner,
		   theFile->c_group,
		   theFile->size,
		   sh_unix_time (theFile->mtime),
		   tmp2);
  SH_FREE(tmp2);

  /* --- Check for links. ---
   */
  if (theFile->c_mode[0] == 'l') 
    {

      linknamebuf = SH_ALLOC(PATH_MAX + 2);

      sl_set_suid();
      linksize    = readlink (theFile->fullpath, linknamebuf, PATH_MAX);
      sl_unset_suid();
      if (linksize < PATH_MAX && linksize >= 0) 
	linknamebuf[linksize] = '\0';
      else 
	linknamebuf[PATH_MAX] = '\0';
      
      if (linksize < 0) 
	{
	  linksize = errno;
	  tmp2 = sh_util_safe_name (theFile->fullpath);
	  sh_error_handle (level, FIL__, __LINE__, linksize, MSG_FI_RDLNK,
			   sh_error_message (linksize), tmp2);
	  SH_FREE(tmp2);
	  SH_FREE(linknamebuf);
	  SL_RETURN((-1),_("sh_unix_getinfo"));
	}
    
    if (linknamebuf[0] == '/') 
      {
	sl_strlcpy (theFile->linkpath, linknamebuf, PATH_MAX+1);
      } 
    else 
      {
	tmp = sh_util_basename(theFile->fullpath);
	sl_strlcpy (theFile->linkpath, 
		    tmp, 
		    PATH_MAX+1);
	SH_FREE(tmp);
	sl_strlcat (theFile->linkpath, 
		    "/", PATH_MAX+1);
	sl_strlcat (theFile->linkpath, 
		    linknamebuf, 
		    PATH_MAX+1);
      }
    
    /* stat the link
     */
    sl_set_suid();
    stat_return = retry_lstat (FIL__, __LINE__, theFile->linkpath, &lbuf); 
    sl_unset_suid();

    /* check for error
     */
    if (stat_return != 0) 
      { 
	stat_return = errno;
	tmp  = sh_util_safe_name (theFile->fullpath);
	tmp2 = sh_util_safe_name (theFile->linkpath);
	if (stat_return != ENOENT) 
	  sh_error_handle (level, FIL__, __LINE__, stat_return, MSG_FI_LSTAT,
			   sh_error_message (stat_return), tmp2);
	else /* a dangling link -- everybody seems to have plenty of them */
	  sh_error_handle ((-1), FIL__, __LINE__, 0, MSG_FI_DLNK,
			   tmp, tmp2);
	theFile->linkisok = BAD;
	SH_FREE(tmp);
	SH_FREE(tmp2);
	SH_FREE(linknamebuf);
	SL_RETURN((-1),_("sh_unix_getinfo"));
      }

    theFile->linkisok = GOOD;
      

    /* --- Determine file type. ---
     */
    sh_unix_getinfo_type (&lbuf, &type, theFile->link_c_mode);
    theFile->type = type;
  
    /* --- Determine permissions. ---
     */
    sh_unix_getinfo_mode (&lbuf, &mode, theFile->link_c_mode);
    theFile->linkmode = lbuf.st_mode;

    /* --- Output the link. ---
     */
    if (theFile->linkisok == GOOD) 
      {
	tmp2 = sh_util_safe_name (linknamebuf);      
	sh_error_handle ((-1), FIL__, __LINE__, 0, MSG_FI_LLNK,
			 theFile->link_c_mode, tmp2);
	SH_FREE(tmp2);
      }
    SH_FREE(linknamebuf);
  }  
  SL_RETURN((0),_("sh_unix_getinfo"));
}

/*  #if defined (SH_WITH_CLIENT) || defined (SH_STANDALONE)  */
#endif

extern int get_the_fd (SL_TICKET ticket);

/* check whether file is locked
 */
int sh_unix_testlock (char * filename)
{
  static        struct stat   buf;
  int           status = 0;

  SL_TICKET     fd;
  char          line_in[128];

  SL_ENTER(_("sh_unix_testlock"));

  if (filename != NULL)
    {
      sl_set_suid();
      status = retry_lstat (FIL__, __LINE__, filename, &buf);
      sl_unset_suid();

      /* no logfile to lock
       */
      if (status < 0)
	SL_RETURN((-1),_("sh_unix_testlock"));
    }

  sl_set_suid();
  status = retry_lstat (FIL__, __LINE__, sh.srvlog.alt, &buf);
  sl_unset_suid();

  /* --- No lock file found, try to lock. ---
   */

  if (status < 0 && errno == ENOENT)
    {
      if (0 == sh_unix_lock ())
	{  
	  sh.flag.islocked = GOOD;
	  SL_RETURN((0),_("sh_unix_testlock"));
	}
    }
  else if (status == 0 && buf.st_size == 0)
    {
      sh.flag.islocked = GOOD;
      sh_unix_unlock ();
      sh.flag.islocked = BAD;
      if (0 == sh_unix_lock ())
	{  
	  sh.flag.islocked = GOOD;
	  SL_RETURN((0),_("sh_unix_testlock"));
	}
    }

  /* --- Check on lock. ---
   */
  
  if (status >= 0)
    fd = sl_open_read (sh.srvlog.alt, SL_YESPRIV);
  else
    fd = -1;

  if (!SL_ISERROR(fd))
    {
      
      /* read the PID in the lock file
       */
      status = sh_unix_getline (fd, line_in, sizeof(line_in)-1);

      /* convert to numeric
       */
      if (status > 0)
	{
	  errno  = 0;
	  status = strtol(line_in, (char **)NULL, 10);
	  if (errno == ERANGE || status <= 0)
	    status = -1;
	}
      sl_close(fd);

      if (status == (int) getpid())
	{
	  sh.flag.islocked = GOOD;
	  SL_RETURN((0),_("sh_unix_testlock"));
	}


      /* --- Check whether the process exists. ---
       */
      if (status > 0)
	{
	  errno  = 0;
	  status = aud_kill (FIL__, __LINE__, status, 0);

	  /* Does not exist, so remove the stale lock
	   * and create a new one.
	   */
	  if (status < 0 && errno == ESRCH)
	    {
	      sh.flag.islocked = GOOD;
	      if (0 != sh_unix_unlock())
		sh.flag.islocked = BAD;
	      else
		{
		  if (0 == sh_unix_lock  ())
		    {
		      sh.flag.islocked = GOOD;
		      SL_RETURN((0),_("sh_unix_testlock"));
		    }
		  sh.flag.islocked = BAD;
		}
	    }
	  else
	    {
	      sh_error_handle ((-1), FIL__, __LINE__, status,
			       MSG_E_SUBGEN, 
			       _("Cannot remove stale PID file, PID may be a running process"),
			       _("sh_unix_testlock"));
	      sh.flag.islocked = BAD;
	    }
	}
    }
  SL_RETURN((-1),_("sh_unix_testlock"));
}

int sh_unix_unlock()
{
  int         error = 0;
  
  SL_ENTER(_("sh_unix_unlock"));

  /* --- Logfile is not locked to us. ---
   */
  if (sh.flag.islocked == BAD) 
    SL_RETURN((-1),_("sh_unix_unlock"));

  /* --- Check whether the directory is secure. ---
   */
  if (0 != tf_trust_check (sh.srvlog.alt, SL_YESPRIV))
    SL_RETURN((-1),_("sh_unix_unlock"));

  /* --- Delete the lock file. --- 
   */
  sl_set_suid(); 
  error = retry_aud_unlink (FIL__, __LINE__, sh.srvlog.alt);
  sl_unset_suid(); 
  
  if (error == 0)
    sh.flag.islocked = BAD; /* not locked anymore */
  else
    {
      error = errno;
      sh_error_handle ((-1), FIL__, __LINE__, error, MSG_E_UNLNK,
		       sh_error_message(error), sh.srvlog.alt);
      SL_RETURN((-1),_("sh_unix_unlock"));
    }
  SL_RETURN((0),_("sh_unix_unlock"));
}

int sh_unix_lock ()
{
  struct stat buf;
  int status;
  int errnum;
  char myPid[64];
  SL_TICKET  fd;


  sl_set_suid(); 
  status = retry_lstat (FIL__, __LINE__, sh.srvlog.alt, &buf);
  sl_unset_suid(); 
  
  SL_ENTER(_("sh_unix_lock"));

  if (0 == status) 
    {
      sh.flag.islocked       = BAD;
      SL_RETURN((-1),_("sh_unix_lock"));
    }

  sprintf (myPid, "%ld\n", (long) getpid());           /* known to fit  */

  fd = sl_open_write (sh.srvlog.alt, SL_YESPRIV);

  if (!SL_ISERROR(fd))
    {
      errnum = sl_write (fd, myPid, sl_strlen(myPid));
      sl_close (fd);

      if (!SL_ISERROR(errnum))
	{
	  sh.flag.islocked = GOOD;
	  SL_RETURN((0),_("sh_unix_lock"));
	}
    }

  TPT((0, FIL__, __LINE__, _("msg=<open pid file failed>\n")));
  sh.flag.islocked       = BAD;
  SL_RETURN((-1),_("sh_unix_lock"));

  /* notreached */
}

/* Test whether file exists, is a character device, allows read
 * access, and is not world writeable.
 */
int sh_unix_file_exists(int fd)
{
  struct stat buf;

  SL_ENTER(_("sh_unix_file_exists"));

  if (retry_fstat(FIL__, __LINE__, fd, &buf) == -1)
    SL_RETURN( (-1), _("sh_unix_file_exists"));
  else if ( S_ISCHR(buf.st_mode) &&  0 != (S_IROTH & buf.st_mode)
	    /*
	     * #if !defined(__CYGWIN32__) && !defined(__CYGWIN__) 
	     * && 0 == (S_IWOTH & buf.st_mode)
	     * #endif
	     */
	    ) 
    SL_RETURN( (0), _("sh_unix_file_exists"));
  else 
    SL_RETURN( (-1), _("sh_unix_file_exists"));
}

static char preq[16];

/* Return the path to the configuration/database file.
 */
char * file_path(char what, char flag)
{
  static int init = 0;

  SL_ENTER(_("file_path"));

  if (init == 0)
    {
      sl_strlcpy(preq, _("REQ_FROM_SERVER"), 16);
      ++init;
    }

  switch (what)
    {

    case 'C':
      if (0 == sl_strncmp (sh.conf.path, preq, 15))
	{
#if defined(SH_WITH_SERVER)
	  if (sh.flag.isserver == S_TRUE && sl_strlen(sh.conf.path) == 15)
	    SL_RETURN( NULL, _("file_path"));
	  if (sh.flag.isserver == S_TRUE)
	    SL_RETURN( &(sh.conf.path[15]), _("file_path"));
#endif
	  if (flag == 'R')
	    SL_RETURN( preq, _("file_path"));
	  if (flag == 'I')
	    {
	      if (sl_strlen(sh.conf.path) == 15)
		SL_RETURN( NULL, _("file_path"));
	      else
		SL_RETURN( &(sh.conf.path[15]), _("file_path"));
	    }
	  SL_RETURN ( preq, _("file_path"));
	}
      else
	SL_RETURN( sh.conf.path, _("file_path"));
      break;

    case 'D':
      if (0 == sl_strncmp (sh.data.path, preq, 15))
	{
	  if (flag == 'R')
	    SL_RETURN( preq, _("file_path"));
	  if (flag == 'W' && sl_strlen(sh.data.path) == 15)
	    SL_RETURN (NULL, _("file_path"));
	  if (flag == 'W')
	    SL_RETURN( &(sh.data.path[15]), _("file_path"));
	}
      else
	SL_RETURN( sh.data.path, _("file_path"));
      break;
	
    default:
      SL_RETURN( NULL, _("file_path"));
      break;
    }

  return NULL; /* notreached */
}

/************************************************/
/************************************************/
/****   Stealth Utilities                    ****/
/************************************************/
/************************************************/
#ifdef SH_STEALTH

void sh_unix_xor_code (char * str, int len)
{
  register int i;

  for (i = 0; i < len; ++i) str[i] ^= (char) XOR_CODE;
  return;
}

#if  !defined(SH_STEALTH_MICRO)

static unsigned long off_data = 0;
static unsigned long max_data = 0;
static int           stealth_init = BAD;

int hideout_hex_block(SL_TICKET fd, unsigned char * str, int len);
unsigned long first_hex_block(SL_TICKET fd, unsigned long * max);

/*
 * --- Get hidden data from a block of hex data. ---
 */
int sh_unix_getline_stealth (SL_TICKET fd, char * str, int len)
{
  int add_off, llen;

  SL_ENTER(_("sh_unix_getline_stealth"));

  /* --- Initialize. ---
   */
  if (stealth_init == BAD)
    {
      off_data = first_hex_block(fd, &max_data);
      if (off_data == 0)
	{
	  dlog(1, FIL__, __LINE__, 
	       _("The stealth config file does not contain any steganographically\nhidden data. This file must be an image file in _uncompressed_\npostscript format.\nTo hide data in it, use:\n   samhain_stealth -s postscript_file orig_config_file\n   mv postscript_file /path/to/config/file\n"));
	  sh_error_handle ((-1), FIL__, __LINE__,  EIO, MSG_P_NODATA,
			   _("Stealth config file."));
	  aud_exit (FIL__, __LINE__, EXIT_FAILURE);
	}
      stealth_init = GOOD;
    }
  
  /* --- Seek to proper position. ---
   */
  sl_seek(fd, off_data);
     
  /* --- Read one line. ---
   */
  add_off   = hideout_hex_block(fd, (unsigned char *) str, len);
  off_data += add_off;


  llen = sl_strlen(str);
  SL_RETURN(llen, _("sh_unix_getline_stealth"));
}

int hideout_hex_block(SL_TICKET fd, unsigned char * str, int len)
{

  register int  i, j, k;
  unsigned char c, e;
  register int  num;
  unsigned char mask[9] = { 0x80, 0x40, 0x20, 0x10, 0x08, 0x04, 0x02, 0x01 };
  unsigned long here   = 0;
  unsigned long retval = 0;

  SL_ENTER(_("hideout_hex_block"));

  i = 0;
  while (i < len)
    {
      for (j = 0; j < 8; ++j)
	{

	  /* --- Get a low byte, modify, read back. --- 
	   */
	  for (k = 0; k < 2; ++k)
	    {
	      /* -- Skip whitespace. ---
	       */
	      c = ' ';
	      do {
		do {
		  num = sl_read (fd, &c, 1);
		} while (num == 0 && errno == EINTR);
		if (num == 0) 
		  SL_RETURN((-1), _("hideout_hex_block"));
		++here; 
	      } while (c == '\n' || c == '\t' || c == '\r' || 
		       c == ' ');
	    }
	  

	  /* --- e is the value of the low byte. ---
	   */
	  e = (unsigned char) sh_util_hexchar( c );
	  if ((e & mask[7]) != 0)  /* bit is set     */
	    str[i] |= mask[j];
	  else                     /* bit is not set */
	    str[i] &= ~mask[j];

	}
      if (str[i] == '\n') break;
      ++i;
    }

  if (i != 0)
    str[i] = '\0';
  else
    str[i+1] = '\0';
  retval += here;

  SL_RETURN(retval, _("hideout_hex_block"));
}

/* --- Get offset of first data block. ---
 */
unsigned long first_hex_block(SL_TICKET fd, unsigned long * max)
{
  int           i;
  register int  num = 1;
  char          c;
  int           nothex = 0;
  unsigned long retval = 0;
  int           this_line = 0;
  char          theline[SH_BUFSIZE];

  SL_ENTER(_("first_hex_block"));

  *max = 0;

  while (1)
    {
      theline[0] = '\0';
      this_line  = 0;
      c          = '\0';
      while (c != '\n' && num > 0)
	{
	  do {
	    num = sl_read (fd, &c, 1);
	  } while (num == 0 && errno == EINTR);
	  if (num > 0) 
	    theline[this_line] = c;
	  else           
	    SL_RETURN((0), _("first_hex_block"));
	  this_line += num;
	}
      theline[this_line] = '\0';
      
      /* not only 'newline' */ 
      if (this_line > 60)
	{
	  nothex  = 0;
	  i       = 0;
	  while (nothex == 0 && i < (this_line-1))
	    {
	      if (! isxdigit((int)theline[i])) nothex = 1;
	      ++i;
	    }
	  if (nothex == 1) retval += this_line;
	}
      else
	{
	  nothex = 1;
	  retval += this_line;
	}

      if (nothex == 0)
	{
	  *max = 0; 
	  do {
	    do {
	      num = sl_read (fd, theline, SH_BUFSIZE);
	    } while (num == 0 && errno == EINTR);
	    for (i = 0; i < num; ++i)
	      { 
		c = theline[i];
		if (c == '\n' || c == '\t' || c == '\r' || c == ' ') 
		  ;
		else if (!isxdigit((int)c))
		  break;
		else
		  *max += 1;
	      }
	  } while (num > 0);

	  *max /= 16;
	  SL_RETURN((retval), _("first_hex_block"));
	}

    }
  SL_RETURN((0), _("first_hex_block"));
}

 /* if !defined(SH_STEALTH_MICRO) */
#endif 

 /* ifdef SH_STEALTH */
#endif


  

#if defined(SCREW_IT_UP) && defined(HAVE_PTRACE) && !defined(sun) && !defined(__sun) && !defined(__sun__)
#include <sys/wait.h>
#include <sys/ptrace.h>

inline
int sh_derr ()
{
  int p1 = 0, p2 = 0;

  p1 = getppid();


#if defined(PTRACE_ATTACH) && defined(PTRACE_DETACH)

  if (0 == (p2 = fork()))
    {
      /* ---  The child.  ---
       */
      p2 = getppid(); 
      if (ptrace(PTRACE_ATTACH, p2, 0, 0) < 0) 
	{
	  ptrace(PTRACE_ATTACH, p1, 0, 0); 
	  kill (p2, 9);
	}

#ifdef HAVE_USLEEP
      usleep(100);
#else
      sleep (1);
#endif

      ptrace(PTRACE_DETACH, p2, 0, 0);

      /* need to send SIGCONT to make sure the parent process gets
       * out of stopped state (one more ugly kludge)
       */
      kill(p2, SIGCONT);
      _exit(EXIT_SUCCESS);
    }
  if (p2 != (-1))
    waitpid (p2, &p1, 0);

#elif defined(PT_ATTACH) && defined(PT_DETACH)

  if (0 == (p2 = fork()))
    {
      /* ---  The child.  ---
       */
      p2 = getppid(); 

#ifdef __HP_UX
      if (ptrace(PT_ATTACH, p2, 0, 0, 0) < 0)
#else 
      if (ptrace(PT_ATTACH, p2, 0, 0) < 0)
#endif

	{

#ifdef __HP_UX
	  ptrace(PT_ATTACH, p1, 0, 0, 0);
#else 
	  ptrace(PT_ATTACH, p1, 0, 0);
#endif

	  kill (p2, 9);
	}
#ifdef HAVE_USLEEP
      usleep(100);
#else
      sleep (1);
#endif

#ifdef __HP_UX
      ptrace(PT_DETACH, p2, 1, 0, 0);
#else 
      ptrace(PT_DETACH,   p2, (char *) 1, 0);
#endif

      /* need to send SIGCONT to make sure the parent process gets
       * out of stopped state (one more ugly kludge)
       */
      kill(p2, SIGCONT);
      _exit(EXIT_SUCCESS);
    }

  if (p2 != (-1))
    waitpid (p2, &p1, 0);

  /* defined(PT_ATTACH) && defined(PT_DETACH) */
#endif

  return 0;
}
  /* if defined(SCREW_IT_UP) && defined(HAVE_PTRACE) */
#else

inline
int sh_derr ()
{
  return 0;
}

#endif




